// Project HPSensorsESP
// 21-Jan-17 This is the one that works! Installed in Daiichi external unit.

#include <ArduinoJson.h>
#include <math.h>
#include <ArduinoOTA.h>
#include <Wire.h>
#include <Ticker.h>
#include "ESPAsyncTCP.h"
#include "ESPAsyncWebServer.h"
#include "HPControllerESP.h"
#include "Weather.h"
#include "sensors.h"
#include "SystemState.h"
#include "debugWeb.h"

#define WIFIPORT 8081
#define ESP8266
#define SPIFFS_ALIGNED_OBJECT_INDEX_TABLES 1

AsyncWebServer server(WIFIPORT);
AsyncEventSource statusEvent("/updates"); // event source (Server-Sent events)
AsyncEventSource weatherEvent("/weather"); // event source (Server-Sent events)
AsyncEventSource debugEvent("/debugMsg"); // event source (Server-Sent events)

Sensors sensors;
SystemState systemState;
EepromUtil eepromUtil;
DebugWeb debugWeb;

const char* ssid = "Terrapico";
const char* password = "5432167890";
const char* hostName = "HPController";
const char* version = "V1.0";

Ticker tickerOSWatch, tmrGetNTP, tmrGetSensors, tmrDefrostTick, tmrMain, tmrSyncTime, tmrTick, tmrDewpoint;
#define OSWATCH_RESET_TIME 30
static unsigned long last_loop;

bool bSyncNTP;
bool bUpdateTime;
bool bGetSensors;
bool bMain;
bool bDewpoint = TRUE;
bool bDefrost = FALSE;

//flag to use from web update to reboot the ESP
bool shouldReboot = false;
long millisNow;

void setup() {
	char result[16];
	byte ipAddress[4];

	Wire.begin(4, 5);							//sda, scl
	Wire.setClockStretchLimit(15000);
	expanderWrite(0xff);						//Set everything off just in case

	Serial.begin(115200);
	delay(100);
	Serial.setDebugOutput(true);

	//do {
	//	setRelay(COMPRESSOR, OFF);
	//	delay(1000);
	//	setRelay(COMPRESSOR, ON);
	//	delay(1000);
	//} while (1);

	//Set weather forecast to 0
	for(int i = 0; i < 7; i++)
		forecast[i] = { 0,0,0,0,0,0 };

	DEBUG_PRINTLN("");
	DEBUG_PRINTLN(version);
	DEBUG_PRINTLN("Setup....");

	systemState.readSettings();				//Read settings from EEPROM

	DEBUG_PRINTLN("Startup wifi");
	WiFi.mode(WIFI_STA);					//WIFI_STA
	WiFi.hostname(hostName);

	WiFi.begin(ssid, password);
	if (WiFi.waitForConnectResult() != WL_CONNECTED) {
		DEBUG_PRINTLN("WiFi Failed! Rebooting...");
		delay(100);
		ESP.restart();
		return;
	}
	DEBUG_PRINTLN("WiFi connected!");
	DEBUG_PRINTST("%d.%d.%d.%d:%d %s\n", WiFi.localIP()[0], WiFi.localIP()[1], WiFi.localIP()[2], WiFi.localIP()[3], WIFIPORT, version);
	systemState.wifiActive = TRUE;

	SPIFFS.begin();
	DEBUG_PRINTLN("SPIFFS started");

	setupWebServer();
	setupOTAUpdate();
	setupNTP();									//Setup NTP time and read time

	systemState.defrostCount = systemState.defrostRunTime * 60;
	systemState.defrostStepCount = systemState.defrostCount;
	millisNow = millis();							//Save current millis in case we need a startup delay
	
	if (systemState.operatingMode == COOL)
		systemState.targetTemperature = systemState.coolTemperature;
	else
		systemState.targetTemperature = systemState.heatTemperature;
	sensors.readTemperatureSensors();

	delay(500);
	getWeather();
	delay(500);
	setupBME();				//Presure and humidity sensors

	//Start timers for system functions
	tmrGetSensors.attach(20, doGetSensors);
	tmrTick.attach(1, doTick);
	tmrMain.attach(4, doMain);
	tmrDewpoint.attach(30, doDewpoint);
	tmrSyncTime.attach(3600, doSyncNTPTime);

	//Set up watchdog timer
	last_loop = millis();
	tickerOSWatch.attach_ms(((OSWATCH_RESET_TIME / 3) * 1000), osWatch);
	DEBUG_PRINTLN("Setup done");
}

//Flags set by timers to run system functions
void doDewpoint() { bDewpoint = TRUE; }
void doMain() {bMain = TRUE; }
void doSyncNTPTime() { bSyncNTP = TRUE; }
void doGetSensors() { bGetSensors = TRUE; }
void doDefrost() { systemState.defrostCount++; bDefrost = TRUE; }
void doTick() { bUpdateTime = TRUE; }

// Main program loop
void loop() {

	ArduinoOTA.handle();
	last_loop = millis();							//Timer for watchdog

	if (bUpdateTime == TRUE) {
		ntpTime = readTime();
		statusEvent.send(ntpTime.c_str(), "timestamp");
		bUpdateTime = FALSE;
	}

	if (bSyncNTP == TRUE) {
		DEBUG_PRINTLN("Update time");
		syncTime();
		getWeather();
		bSyncNTP = FALSE;
	}

	if (bGetSensors == TRUE) {
		DEBUG_PRINTLN("Read sensors");
		sensors.readTemperatureSensors();
		readBMSensors();
		bGetSensors = FALSE;
	}

	if (shouldReboot) {
		DEBUG_PRINTLN("Rebooting...");
		delay(100);
		ESP.restart();
	}

	if (bDefrost == TRUE)
	{
		defrostCheck();
		bDefrost = FALSE;
	}

	if (bDewpoint == TRUE)
	{
		DEBUG_PRINTLN("Read dewpoint");
		getDewpoint();
		bDewpoint = FALSE;
	}

	if (bMain == TRUE)
	{
		checkPump();
		if (systemState.startupDelay == FALSE)
			threadMain();
		else
		{
			DEBUG_PRINTST("delay:%ld\n", (millisNow + 120000) - (millisNow + millis()));
			if (millisNow + millis() > millisNow + 120000)
				systemState.startupDelay = FALSE;
		}
		if (systemState.dataDirty == TRUE)
		{
			sendWeatherJson();						//Send weather information
			sendRoomJson();							//Send room (dewpoint) information
			systemState.sendSystemStatus();			//Send json with system status to web browser
			systemState.dataDirty = FALSE;
		}
		bMain = FALSE;
	}
}

//Main logic
void threadMain() {

	//modeChange - need to implement
	checkWebUpdate();

	//If it is off turn everything off
	if ((systemState.operatingState == ON) && (systemState.onOff == OFF))
	{
		DEBUG_PRINTLN("Turning OFF");
		setRelay(COMPRESSOR, OFF);
		setRelay(FAN, OFF);
		setRelay(VALVE, OFF);
		systemState.defrostFlag = FALSE;
		systemState.setOperatingState(OFF);
	}

	if (((systemState.operatingState == OFF) && (systemState.onOff == OFF)) || (systemState.defrostFlag == TRUE))
		return;

	//If compressor temperature exceeds maximum turn the machine off
	if (TEMP_COMPRESSOR > systemState.maxCompressorTemperature)
	{
		systemState.onOff = OFF;
		DEBUG_PRINTLN("Compressor overheat - turn off");
	}

	//Machine startup
	if (systemState.operatingMode == HEAT) {
		if (systemState.operatingState == ON) {
			if (TEMP_ACTUAL >= systemState.heatTemperature) {
				DEBUG_PRINTLN("Target reached turning off - HEAT");
				//If we have achieved the target temperature then turn everything off
				setRelay(VALVE, OFF);
				setRelay(FAN, OFF);
				setRelay(COMPRESSOR, OFF);
				systemState.setOperatingState(OFF);
			}

			//Check to see if we need to do a defrost. 
			if (TEMP_EVAPORATOR < systemState.defrostTemperature) {
				if (millis() - lastDefrost >= (E_DEFROSTDELAY * 1000))
				{
					if (systemState.defrostFlag == FALSE)
					{
						DEBUG_PRINTLN("Start defrost cycle");
						systemState.defrostCount = 0;
						systemState.defrostFlag = TRUE;
						tmrDefrostTick.attach(1, doDefrost);
					}
				}
			}
		}
		else
		{
			//Operational modes for heating. Override hysterisis if just turned on
			int hys;
			if (systemState.onFlag == TRUE) {
				systemState.onFlag = FALSE;
				hys = 0;
			}
			else
				hys = systemState.hysterisis;

			if (hys != 0)
				DEBUG_PRINTLN("Wait on hysterisis drop");
			if (TEMP_ACTUAL <= systemState.heatTemperature - hys) {
				DEBUG_PRINTLN("Turning on - HEAT");
				setRelay(FAN, ON);
				delay(500);
				setRelay(VALVE, ON);
				delay(5000);
				setRelay(COMPRESSOR, ON);
				systemState.setOperatingState(ON);
			}
		}
	}

	//For cooling
	if (systemState.operatingMode == COOL) {
		if (systemState.operatingState == ON) {
			//Operational modes for cooling
			if (TEMP_ACTUAL <= systemState.coolTemperature) {
				//If we have achieved the target temperature then turn everything off
				DEBUG_PRINTLN("Target reached turning off - COOL"); 
				setRelay(VALVE, OFF);
				setRelay(FAN, OFF);
				setRelay(COMPRESSOR, OFF);
				systemState.setOperatingState(OFF);
			}
		}
		else
		{
			//Operational modes for cooling
			if (TEMP_ACTUAL >= systemState.coolTemperature + systemState.hysterisis) {
				DEBUG_PRINTLN("Turning on - COOL"); 
				setRelay(VALVE, OFF);
				setRelay(FAN, ON);
				delay(5000);
				setRelay(COMPRESSOR, ON);
				systemState.setOperatingState(ON);
			}
		}
	}

	setRelay(0, TRUE);					//Update relays every cycle in case the update didn't work for some reason
}

void ICACHE_RAM_ATTR osWatch(void) {
	unsigned long t = millis();
	unsigned long last_run = abs(t - last_loop);
	if (last_run >= (OSWATCH_RESET_TIME * 1000)) {
		// save the hit here to eeprom or to rtc memory if needed
		expanderWrite(7);						//Everything off
		ESP.restart();  // normal reboot 
						//ESP.reset();  // hard reset
	}
}

//Check circulating pump and turn on/off as necessary
void checkPump() {
	//Control for circulating pump
	if ((systemState.onOff == ON) && (systemState.pumpState == OFF))
	{
		DEBUG_PRINTLN("Turning pump on");
		setRelay(PUMP, ON);
	}
	if ((systemState.onOff == OFF) && (systemState.pumpState == ON))
	{
		DEBUG_PRINTLN("Turning pump off");
		setRelay(PUMP, OFF);
	}
}

// Turn on/off relays and preserve state flag
void setRelay(int relay, bool setting)
{
	char temp[20];
	byte data = 0xff;

	switch (relay) {
	case COMPRESSOR:
		if (systemState.compressorState != setting) {
			systemState.compressorState = setting;
		}
		break;
	case VALVE:
		if (systemState.valveState != setting) {
			systemState.valveState = setting;
		}
		break;
	case FAN:
		if (systemState.fanState != setting) {
			systemState.fanState = setting;
		}
		break;
	case PUMP:
		if (systemState.pumpState != setting) {
			systemState.pumpState = setting;
		}
		break;
	}

	//Build the value to send to the PCF8574
	if (systemState.compressorState == ON)
		data &= COMPRESSOR;
	if (systemState.valveState == ON)
		data &= VALVE;
	if (systemState.fanState == ON)
		data &= FAN;
	if (systemState.pumpState == ON)
		data &= PUMP;

	DEBUG_PRINTST("Setting relays %d\n", data);
	expanderWrite(data);

	//Send update if we are really changing a state
	if (relay != 0) {
		sprintf(temp, "C:%s F:%s V:%s P:%s", oState[!systemState.compressorState], oState[systemState.fanState], oState[systemState.valveState], oState[systemState.pumpState]);
		statusEvent.send(temp, "relays");
	}
	delay(100);
}

//Check if new value diffrent from old value and if so send the updated value.
//Saves unnecessary network traffic
int checkSend(int current, int previous, char* tag)
{
	char temp[20];

	if (previous != current)
	{
		sprintf(temp, "%d", current);
		//statusEvent.send(temp, tag);
		DEBUG_PRINTST("Send update for %s: %s\n", tag, temp);
		systemState.dataDirty = TRUE;
	}
	return current;
}

bool checkSend(bool current, bool previous, char* tag)
{
	char temp[20];

	if (previous != current)
	{
		sprintf(temp, "%d", current);
		statusEvent.send(temp, tag);
		DEBUG_PRINTST("Send update for %s: %s\n", tag, temp);
		delay(300);
	}
	return current;
}

void checkWebUpdate() {
	char temp[20];

	p_temp_compressor = checkSend(TEMP_COMPRESSOR, p_temp_compressor, "compressor");
	p_defrostCount = checkSend(systemState.defrostStepCount - systemState.defrostCount, p_defrostCount, "defrost");
	p_temp_evapmid = checkSend(TEMP_EVAPMID, p_temp_evapmid, "midevap");
	p_temp_external = checkSend(TEMP_EXTERNAL, p_temp_external, "external");
	p_temp_actual = checkSend(TEMP_ACTUAL, p_temp_actual, "actual");
	p_targetTemperature = checkSend(systemState.targetTemperature, p_targetTemperature, "target");
	p_temp_evaporator = checkSend(TEMP_EVAPORATOR, p_temp_evaporator, "evaporator");

	if ((p_compressorState != systemState.compressorState) || (p_fanState != systemState.fanState) || (p_valveState != systemState.valveState))
	{
		sprintf(temp, "C:%s F:%s V:%s", oState[systemState.compressorState], oState[systemState.fanState], oState[systemState.valveState]);
		statusEvent.send(temp, "relays");
		p_compressorState = systemState.compressorState;
		p_fanState = systemState.fanState;
		p_valveState = systemState.valveState;
	}

	if (p_operatingMode != systemState.operatingMode)
	{
		sprintf(temp, "%s", oMode[systemState.operatingMode]);
		statusEvent.send(temp, "mode");
		p_operatingMode = systemState.operatingMode;
	}

	if (p_onOff != systemState.onOff)
	{
		sprintf(temp, "%s", oState[systemState.onOff]);
		statusEvent.send(temp, "onoff");
		p_onOff = systemState.onOff;
	}
}

//PCF8574 I2C Expander
void expanderWrite(byte data)
{
	Wire.beginTransmission(PCF8574);
	Wire.write(data);
	Wire.endTransmission();
}

// dewPoint function NOAA
// reference (1) : http://wahiduddin.net/calc/density_algorithms.htm
// reference (2) : http://www.colorado.edu/geography/weather_station/Geog_site/about.htm
//
double dewPoint(double celsius, double humidity)
{
	// (1) Saturation Vapor Pressure = ESGG(T)
	double RATIO = 373.15 / (273.15 + celsius);
	double RHS = -7.90298 * (RATIO - 1);
	RHS += 5.02808 * log10(RATIO);
	RHS += -1.3816e-7 * (pow(10, (11.344 * (1 - 1 / RATIO))) - 1);
	RHS += 8.1328e-3 * (pow(10, (-3.49149 * (RATIO - 1))) - 1);
	RHS += log10(1013.246);

	// factor -3 is to adjust units - Vapor Pressure SVP * humidity
	double VP = pow(10, RHS - 3) * humidity;

	// (2) DEWPOINT = F(Vapor Pressure)
	double T = log(VP / 0.61078);   // temp var
	return (241.88 * T) / (17.558 - T);
}

//Debug out system status
void debugStateOutput(void) {
	int desiredTemp;

	desiredTemp = (systemState.operatingMode == HEAT) ? systemState.heatTemperature : systemState.coolTemperature;
	DEBUG_PRINTST("Power:%s<br>Operating Mode:%s<br>operatingState:%s<br>desiredTemp=%d", oState[systemState.onOff], oMode[systemState.operatingMode], oState[systemState.operatingState], desiredTemp);
	if (systemState.startupDelay == TRUE)
		DEBUG_PRINTLN("Startup delay...");
	DEBUG_PRINTST("defrostFlag: %s\n", systemState.defrostFlag == true?"True":"False");
}
