#include "SystemState.h"
#include <EEPROM.h>
#include "EepromUtil.h"
#include "debugWeb.h"
#include "sensors.h"

SystemState::SystemState() {
	readSettings();
}

void SystemState::readSettings(){
	char result[16];

	DEBUG_PRINTLN("Read settings");
	EEPROM.begin(512);
	eepromUtil.eeprom_read_string(E_INIT, result, 5);
	if (strcmp(result, "INIT") != 0) {
		//EEPROM not initialised so set default values
		DEBUG_PRINTLN("Set EEPROM values to default");
		resetParameters();
	}

	defrostTemperature = eepromUtil.eeprom_read_int(E_DEFROSTTEMP) - DEFROST_OFFSET;		//Temperature to run defrost cycle
	defrostRunTime = eepromUtil.eeprom_read_int(E_DEFROSTTIME);								//Defrost cycle time (minutes)
	defrostUpperTemperature = eepromUtil.eeprom_read_int(E_DEFROSTMAXTEMP);					//Defrost Maximum temperature (temperature to stop defrost, overrides timer)
	defrostStartDelay = eepromUtil.eeprom_read_int(E_DEFROSTDELAY);
	maxCompressorTemperature = eepromUtil.eeprom_read_int(E_MAXCOMPTEMP);					//Maximum compressor temperature for safety
	heatTemperature = eepromUtil.eeprom_read_int(E_HEATTEMP);								//Desired room temperature
	coolTemperature = eepromUtil.eeprom_read_int(E_COOLTEMP);								//Desired room temperature
	hysterisis = eepromUtil.eeprom_read_int(E_HYSTERISIS);									//Hysterisis
	systemState.operatingMode = eepromUtil.eeprom_read_int(E_OPERATINGMODE);				//Operating mode
	systemState.setRoomTemperature = eepromUtil.eeprom_read_int(E_ROOMTEMP);				//Room temperature
	systemState.onOff = eepromUtil.eeprom_read_int(E_ONOFF);
	eepromUtil.eeprom_read_string(E_USER, www_username, sizeof(www_username));
	eepromUtil.eeprom_read_string(E_PASS, www_password, sizeof(www_password));
	port = eepromUtil.eeprom_read_int(E_PORT);
	systemState.operatingState = eepromUtil.eeprom_read_int(E_LASTSYSTEMSTATE);				//Last operating state. If running then we need a delay on startup
	if (systemState.operatingState == ON)
	{
		startupDelay = TRUE;
		setOperatingState(OFF);
		DEBUG_PRINTLN("StartupDelay=TRUE");
	}
	debugOutParameters();

}

//Reset all parameters to default values
void SystemState::resetParameters() {
	EEPROM.begin(512);
	eepromUtil.eeprom_erase_all(0x00);
	EEPROM.write(E_DEFROSTDELAY, 5);
	EEPROM.write(E_DEFROSTTIME, 5);
	EEPROM.write(E_DEFROSTTEMP, -6 + DEFROST_OFFSET);
	EEPROM.write(E_DEFROSTMAXTEMP, 10);
	EEPROM.write(E_MAXCOMPTEMP, 125);
	EEPROM.write(E_HEATTEMP, 34);
	EEPROM.write(E_COOLTEMP, 15);
	EEPROM.write(E_HYSTERISIS, 5);
	EEPROM.write(E_ROOMTEMP, 23);
	EEPROM.write(E_OPERATINGMODE, HEAT);
	EEPROM.write(E_LASTSYSTEMSTATE, OFF);
	EEPROM.write(E_ONOFF, OFF);
	EEPROM.write(E_PORT, WIFIPORT);
	eepromUtil.eeprom_write_string(E_USER, "admin");
	eepromUtil.eeprom_write_string(E_PASS, "admin");
	eepromUtil.eeprom_write_string(E_INIT, "INIT");
	EEPROM.commit();
	//eeprom_serial_dump_xy(0, E_PASS + 20);
	systemState.readSettings();
}

void SystemState::setOperatingState(bool state) {
	systemState.operatingState = state;
	EEPROM.write(E_LASTSYSTEMSTATE, state);
	EEPROM.commit();
}

void SystemState::debugOutParameters() {
	if (systemState.wifiActive == FALSE)
		return;
	delay(100);

	String jsonString = getJsonSettings();
	if (systemState.wifiActive == TRUE) {
		debugEvent.send(jsonString.c_str(), "debug");
	}
}

//Return a string with the system parameters as a json encapsulated in a function for processing on the browser
String SystemState::getJsonSettings() {

	// Add values in the object
	DynamicJsonBuffer jsonBuffer;
	JsonObject& root = jsonBuffer.createObject();
	root["defrostTemperature"] = systemState.defrostTemperature;
	root["defrostRunTime"] = systemState.defrostRunTime;
	root["defrostUpperTemperature"] = systemState.defrostUpperTemperature;
	root["defrostStartDelay"] = systemState.defrostStartDelay;
	root["maxCompressorTemperature"] = systemState.maxCompressorTemperature;
	root["heatTemperature"] = systemState.heatTemperature;
	root["coolTemperature"] = systemState.coolTemperature;
	root["hysterisis"] = systemState.hysterisis;
	root["onOff"] = systemState.onOff;
	root["operatingMode"] = systemState.operatingMode;
	root["operatingState"] = systemState.operatingState;
	root["setRoomTemperature"] = systemState.setRoomTemperature;

	root.prettyPrintTo(Serial);

	String jsonString;
	root.printTo(jsonString);
	//String reply = "statusJson(" + jsonString + ");";
	return jsonString;
}

void SystemState::sendSystemStatus() {
	if (systemState.wifiActive == FALSE)
		return;

	String jsonString = getJsonStatus();
	statusEvent.send(jsonString.c_str(), "statusJson");
}

//Return a string with the system status as a json encapsulated in a function for processing on the browser
String SystemState::getJsonStatus() {
	char relays[30];
	char tempHum[30];

	// Add values in the object
	DynamicJsonBuffer jsonBuffer;
	JsonObject& root = jsonBuffer.createObject();
	root["compressor"] = TEMP_COMPRESSOR;
	root["evaporator"] = TEMP_EVAPORATOR;
	root["midevap"] = TEMP_EVAPMID;
	root["actual"] = TEMP_ACTUAL;
	root["target"] = systemState.targetTemperature;
	root["defrostRunTime"] = systemState.defrostCount;
	sprintf(relays, "C:%s F:%s V:%s P:%s", oState[systemState.compressorState], oState[systemState.fanState], oState[systemState.valveState], oState[systemState.pumpState]);
	root["relays"] = relays;
	root["defrostStep"] = systemState.defrostStep;
	root["lastDefrost"] = systemState.lastDefrost;
	root["onoff"] = oState[systemState.onOff];
	root["mode"] = oMode[systemState.operatingMode];
	root["pressure"] = systemState.pressure;
	root["humidity"] = systemState.humidity;
	root["ambient"] = systemState.ambientTemperature;
	int a = round(systemState.ambientTemperature);
	int b = round(systemState.humidity);
	sprintf(tempHum, "%dC (%dC %d%%)", TEMP_EXTERNAL, a, b);
	root["external"] = tempHum;

	root.prettyPrintTo(Serial);

	String jsonString;
	root.printTo(jsonString);
	//String reply = "valuesJson(" + jsonString + ");";
	return jsonString;
}
