/*

Udp NTP Client

Get the time from a Network Time Protocol (NTP) time server
Demonstrates use of UDP sendPacket and ReceivePacket
For more on NTP time servers and the messages needed to communicate with them,
see http://en.wikipedia.org/wiki/Network_Time_Protocol

created 4 Sep 2010
by Michael Margolis
modified 9 Apr 2012
by Tom Igoe
updated for the ESP8266 12 Apr 2015
by Ivan Grokhotkov

This code is in the public domain.

//#include <TimeArd.h>
//#include <TimeLib.h>
//#include <Time.h>

*/

//#define NOTIME

//#include <TimeArd.h>
#ifndef NOTIME
#include <Timezone.h>    //https://github.com/JChristensen/Timezone
#include "TimeLib.h"
#endif

#include <ESP8266WiFi.h>
#include <WiFiUdp.h>

//Spain
#ifndef NOTIME
TimeChangeRule myDST = { "DST", Last, Sun, Mar, 2, +120 };		//Daylight time = UTC - 1 hours
TimeChangeRule mySTD = { "CET", Last, Sun, Oct, 3, +60 };     //Standard time = UTC - 2 hours
Timezone myTZ(myDST, mySTD);

//If TimeChangeRules are already stored in EEPROM, comment out the three
//lines above and uncomment the line below.
//Timezone myTZ(100);       //assumes rules stored at EEPROM address 100

TimeChangeRule *tcr;        //pointer to the time change rule, use to get TZ abbrev
#endif
time_t utc, local;

unsigned int localPort = 2390;				// local port to listen for UDP packets
IPAddress timeServerIP;						// time.nist.gov NTP server address
const char* ntpServerName = "pool.ntp.org";
const int NTP_PACKET_SIZE = 48;				// NTP time stamp is in the first 48 bytes of the message
byte packetBuffer[NTP_PACKET_SIZE];			//buffer to hold incoming and outgoing packets								
WiFiUDP udp;								// A UDP instance to let us send and receive packets over UDP

//Setup NTP and set the time to it
void setupNTP()
{
	Serial.println("Starting UDP");
	udp.begin(localPort);
	ntpTime = syncTime();
}

String readTime()
{
	utc++;
	return printTime(utc);
}

String syncTime()
{
	int retry;

	time_t newTime;

	for (retry = 0; retry < 5; retry++) {
		newTime = getNtpTime();
		if (newTime != 0) {
			utc = newTime;
			return printTime(utc);
		}
		DEBUG_PRINTLN("Update time failed");
		delay(5000);
	}
	return("");
}

time_t getNtpTime()
{
	const unsigned long seventyYears = 2208988800UL;
	unsigned long secsSince1900;
	char buf[5];

	//get a random server from the pool
	WiFi.hostByName(ntpServerName, timeServerIP);

	sendNTPpacket(timeServerIP); // send an NTP packet to a time server
								 // wait to see if a reply is available
	delay(1000);

	int cb = udp.parsePacket();
	if (!cb) {
		Serial.println("no packet yet");
	}
	else {
		//Serial.print("packet received, length=");
		//Serial.println(cb);
		// We've received a packet, read the data from it
		udp.read(packetBuffer, NTP_PACKET_SIZE); // read the packet into the buffer

												 //The timestamp starts at byte 40 of the received packet and is four bytes, or two words, long. 
												 //First, esxtract the two words:
		unsigned long highWord = word(packetBuffer[40], packetBuffer[41]);
		unsigned long lowWord = word(packetBuffer[42], packetBuffer[43]);
		// combine the four bytes (two words) into a long integer this is NTP time (seconds since Jan 1 1900):
		secsSince1900 = highWord << 16 | lowWord;

		// now convert NTP time into everyday time:
		// Unix time starts on Jan 1 1970. In seconds, that's 2208988800:
		// subtract seventy years:
		return secsSince1900 - seventyYears;;
	}
	return 0;
}

//Function to print time with time zone
String printTime(time_t t)
{
	String timeStr = "";
	char buf[2];
	char temp[10];

#ifndef NOTIME
	local = myTZ.toLocal(t, &tcr);
	sprintf(temp, "%02d:%02d:%02d", hour(local), minute(local), second(local));
#endif
	timeStr += temp;
	return timeStr;
}

// send an NTP request to the time server at the given address
void sendNTPpacket(IPAddress& address)
{
	//Serial.println("sending NTP packet...");
	// set all bytes in the buffer to 0
	memset(packetBuffer, 0, NTP_PACKET_SIZE);
	// Initialize values needed to form NTP request
	packetBuffer[0] = 0b11100011;   // LI, Version, Mode
	packetBuffer[1] = 0;     // Stratum, or type of clock
	packetBuffer[2] = 6;     // Polling Interval
	packetBuffer[3] = 0xEC;  // Peer Clock Precision
							 // 8 bytes of zero for Root Delay & Root Dispersion
	packetBuffer[12] = 49;
	packetBuffer[13] = 0x4E;
	packetBuffer[14] = 49;
	packetBuffer[15] = 52;

	// all NTP fields have been given values, now you can send a packet requesting a timestamp:
	udp.beginPacket(address, 123); //NTP requests are to port 123
	udp.write(packetBuffer, NTP_PACKET_SIZE);
	udp.endPacket();
}

